import calendar
import datetime
from urllib.parse import urlencode
from urllib.request import Request, build_opener
from urllib.error import HTTPError
import json
from pyconnectedcars.Exceptions import ConnectedCarsException


class Connection(object):
    """Connection to Tesla Motors API"""
    def __init__(self, email, password, baseurl=""):
        """Initialize connection object"""
        self.user_agent = 'okhttp/3.6.0'
        if not baseurl:
            self.baseurl = 'https://skoda.connectedcars.dk'
        self.api = '/api/graphql'
        self._query = "\n    mutation RootMutationType($email: String, $password: String) {\n      user: login(email: $email, password: $password) {\n        ...userFields\n      }\n    }\n  \n\n\n    fragment userFields on User {\n      \n  id,\n  firstname,\n  lastname,\n  mobile,\n  email,\n  lang,\n  onboardingFinished,\n  token,\n  cars {\n    id,\n    vin,\n    locationName,\n    name,\n    lat,\n    long,\n    fuelLevel,\n    fuelLevelLiter,\n    fuelLevelUpdatedAt,\n    lockedState,\n    lockedStateUpdatedAt,\n    systemsAreOk,\n    oilLevelIsOk,\n    tirePressureIsOk,\n    batteryChargeIsOk,\n    odometer,\n    imageFilename,\n    updatedAt,\n    service {\n      nextServiceInKm,\n      nextServiceInDays\n    },\n    licensePlates {\n      id,\n      licensePlate,\n      createdAt\n    },\n    lamps(listLampStates: true, source: USER) {\n      type,\n      color,\n      frequency,\n      enabled,\n      source,\n      time\n    },\n    incidents(status: ON, dismissed: false, limit: 1000) {\n      id,\n      rule,\n      system {\n        key,\n        headerDanish,\n        headerEnglish,\n        nameDanish,\n        nameEnglish\n      },\n      recommendation {\n        key,\n        descriptionEnglish,\n        descriptionDanish\n      },\n      startTime,\n      context {\n        ... on CarIncidentServiceReminderContext {\n          serviceDate\n        }\n      }\n    }\n  },\n  workshop {\n    id,\n    dealernumber,\n    dealername,\n    phone,\n    bookingurl\n  }\n\n    }\n  "
        self.auth_vars = {
            "email": email,
            "password": password
        }

    def get_data(self):
        """Utility command to get data from API"""
        payload = {
            "query": self._query,
            "variables": self.auth_vars
        }

        return self.post(json.dumps(payload))

    def post(self, data=""):
        """Utility command to post data to API"""
        headers = {
            'Content-Type': 'application/json',
            "User-Agent": self.user_agent
        }
        return self.__open(
            self.api, headers=headers, req_data=data)

    def __open(self, url, headers={}, req_data=None):
        """Raw urlopen command"""
        req = Request(
            "%s%s" % (self.baseurl, url),
            data=req_data.encode('utf8'),
            headers=headers
        )

        opener = build_opener()

        resp = opener.open(req)
        charset = resp.info().get('charset', 'utf-8')
        data = json.loads(resp.read().decode(charset))
        opener.close()
        # TODO: Parse errors, because why use HTTP error codes...
        print(data)
        if data.get('errors'):
            raise ConnectedCarsException(
                data['errors'][0]['message'], data['errors'])
        return data



import time
from multiprocessing import RLock
from pyconnectedcars.connection import Connection
from pyconnectedcars.Fuel import Fuel
from pyconnectedcars.Lock import Lock
from pyconnectedcars.BinarySensor import SystemsAreOkSensor, OilLevelIsOkSensor, TirePressureIsOkSensor, BatteryChargeIsOkSensor
from pyconnectedcars.GPS import GPS, Odometer


class Controller:
    def __init__(self, email, password, update_interval):
        self.__connection = Connection(email, password)
        self.__vehicles = []
        self.update_interval = update_interval
        self.__car = {}
        self.__last_update_time = 0
        self.__lock = RLock()
        self.update()
        for car_id, car in self.__car.items():
            self.__vehicles.append(Fuel(car, self))
            self.__vehicles.append(Lock(car, self))
            self.__vehicles.append(GPS(car, self))
            self.__vehicles.append(Odometer(car, self))
            self.__vehicles.append(SystemsAreOkSensor(car, self))
            self.__vehicles.append(OilLevelIsOkSensor(car, self))
            self.__vehicles.append(TirePressureIsOkSensor(car, self))
            self.__vehicles.append(BatteryChargeIsOkSensor(car, self))

    def get(self):
        self.__last_update_time = time.time()
        return self.__connection.get_data()

    def list_vehicles(self):
        return self.__vehicles

    def update(self):
        cur_time = time.time()
        with self.__lock:
            if cur_time - self.__last_update_time > self.update_interval:
                data = self.get()
                if data and data['data']:
                    for car_data in data['data']['user']['cars']:
                        self.__car[car_data['id']] = car_data
                else:
                    self.__car = {}

    def get_car_params(self, car_id):
        return self.__car[car_id]



from pyconnectedcars.vehicle import VehicleDevice
import dateutil.parser as parser


class Lock(VehicleDevice):
    def __init__(self, data, controller):
        super().__init__(data, controller)
        self.__lock_state = False

        self.last_updated = 0

        self.type = 'door lock'
        self.hass_type = 'lock'
        self.name = self._name()
        self.uniq_name = self._uniq_name()
        self.bin_type = 0x7
        self.update()

    def update(self):
        self._controller.update()
        data = self._controller.get_car_params(self._id)
        if data:
            if data['lockedState'] == 'UNLOCKED':
                self.__lock_state = False
            else:
                self.__lock_state = True
            self.last_updated = parser.parse(data['fuelLevelUpdatedAt'])

    def is_locked(self):
        return self.__lock_state

    @staticmethod
    def has_battery():
        return False



class VehicleDevice:
    def __init__(self, data, controller):
        self._id = data['id']
        self._vin = data['vin']
        self._raw_name = data['name']
        self._controller = controller
        self.should_poll = True

    def _name(self):
        return "{} {}".format(self._raw_name, self.type)

    def _uniq_name(self):
        return '{} {} {}'.format(self._raw_name, self._vin, self.type)



from pyconnectedcars.vehicle import VehicleDevice
import dateutil.parser


class GPS(VehicleDevice):
    def __init__(self, data, controller):
        super().__init__(data, controller)
        self.__longitude = 0
        self.__latitude = 0
        self.__location = {}

        self.last_seen = 0
        self.last_updated = 0

        self.type = 'location tracker'
        self.hass_type = 'devices_tracker'
        self.bin_type = 0x6
        self.name = self._name()
        self.uniq_name = self._uniq_name()
        self.update()

    def get_location(self):
        return self.__location

    def update(self):
        self._controller.update()
        data = self._controller.get_car_params(self._id)
        if data:
            self.__longitude = data['long']
            self.__latitude = data['lat']
            self.last_updated = dateutil.parser.parse(data['updatedAt'])
        if self.__longitude and self.__latitude:
            self.__location = {'longitude': self.__longitude,
                               'latitude': self.__latitude}

    @staticmethod
    def has_battery():
        return False


class Odometer(VehicleDevice):
    def __init__(self, data, controller):
        super().__init__(data, controller)
        self.__odometer = 0
        self.last_updated = 0

        self.type = 'mileage sensor'
        self.measurement = 'LENGTH_KILOMETERS'
        self.hass_type = 'sensor'
        self.name = self._name()
        self.uniq_name = self._uniq_name()
        self.bin_type = 0xB
        self.update()
        self.__rated = True

    def update(self):
        self._controller.update()
        data = self._controller.get_car_params(self._id)
        if data:
            self.__odometer = data['odometer']
            self.last_updated = dateutil.parser.parse(data['updatedAt'])

    @staticmethod
    def has_battery():
        return False

    def get_value(self):
        return round(self.__odometer, 1)



from pyconnectedcars.vehicle import VehicleDevice
import dateutil.parser


class SystemsAreOkSensor(VehicleDevice):
    def __init__(self, data, controller):
        super().__init__(data, controller)
        self.__state = False
        self.lamps = []
        self.indicators = []

        self.last_updated = None

        self.type = 'system ok sensor'
        self.hass_type = 'binary_sensor'
        self.name = self._name()
        self.uniq_name = self._uniq_name()
        self.bin_type = 0x1
        self.update()

    def update(self):
        self._controller.update()
        data = self._controller.get_car_params(self._id)
        if data:
            self.__state = data['systemsAreOk']
            self.lamps = data['lamps']
            self.indicators = data['incidents']
            self.last_updated = dateutil.parser.parse(data['fuelLevelUpdatedAt'])

    def get_value(self):
        return self.__state

    @staticmethod
    def has_battery():
        return False


class OilLevelIsOkSensor(VehicleDevice):
    def __init__(self, data, controller):
        super().__init__(data, controller)
        self.__state = False

        self.last_updated = None

        self.type = 'oil level sensor'
        self.hass_type = 'binary_sensor'
        self.name = self._name()
        self.uniq_name = self._uniq_name()
        self.bin_type = 0x1
        self.update()

    def update(self):
        self._controller.update()
        data = self._controller.get_car_params(self._id)
        if data:
            self.__state = data['oilLevelIsOk']
            self.last_updated = dateutil.parser.parse(data['fuelLevelUpdatedAt'])

    def get_value(self):
        return self.__state

    @staticmethod
    def has_battery():
        return False


class TirePressureIsOkSensor(VehicleDevice):
    def __init__(self, data, controller):
        super().__init__(data, controller)
        self.__state = False

        self.last_updated = None

        self.type = 'tire pressure sensor'
        self.hass_type = 'binary_sensor'
        self.name = self._name()
        self.uniq_name = self._uniq_name()
        self.bin_type = 0x1
        self.update()

    def update(self):
        self._controller.update()
        data = self._controller.get_car_params(self._id)
        if data:
            self.__state = data['tirePressureIsOk']
            self.last_updated = dateutil.parser.parse(data['fuelLevelUpdatedAt'])

    def get_value(self):
        return self.__state

    @staticmethod
    def has_battery():
        return False


class BatteryChargeIsOkSensor(VehicleDevice):
    def __init__(self, data, controller):
        super().__init__(data, controller)
        self.__state = False

        self.last_updated = None

        self.type = 'battery charge sensor'
        self.hass_type = 'binary_sensor'
        self.name = self._name()
        self.uniq_name = self._uniq_name()
        self.bin_type = 0x1
        self.update()

    def update(self):
        self._controller.update()
        data = self._controller.get_car_params(self._id)
        if data:
            self.__state = data['batteryChargeIsOk']
            self.last_updated = dateutil.parser.parse(data['fuelLevelUpdatedAt'])

    def get_value(self):
        return self.__state

    @staticmethod
    def has_battery():
        return False



from pyconnectedcars.connection import Connection
from pyconnectedcars.Fuel import Fuel
from pyconnectedcars.Lock import Lock
from pyconnectedcars.BinarySensor import (SystemsAreOkSensor, OilLevelIsOkSensor, TirePressureIsOkSensor, BatteryChargeIsOkSensor)
from pyconnectedcars.GPS import GPS, Odometer
from pyconnectedcars.controller import Controller
